"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("./utils");
const cells_1 = require("@jupyterlab/cells");
class ContextManager {
    static context() {
        if (!ContextManager._context) {
            if (ContextManager.is_lab())
                ContextManager._context = new LabContext();
            else if (ContextManager.is_notebook())
                ContextManager._context = new NotebookContext();
            else
                ContextManager._context = new EmbedContext();
        }
        return ContextManager._context;
    }
    /**
     * Determine is this is running in JupyterLab
     *
      * @returns {boolean}
     */
    static is_lab() {
        // The assumption is that only JupyterLab will provide an initialized NotebookTracker object
        // If a better way to detect the environment becomes available, use that instead
        return !!ContextManager.notebook_tracker;
        // A previous implementation replied on checking the DOM
        // return !!document.querySelector("#main.jp-LabShell")
    }
    /**
     * Determine if this is running in the classic Jupyter Notebook
     *
     * @returns {boolean}
     */
    static is_notebook() {
        // Check if the Jupyter global variable has been defined
        return !!(window['Jupyter']);
    }
}
exports.ContextManager = ContextManager;
/**
 * Abstract base class representing the context in which the widgets are running.
 * Contexts include: JupyterLab, Jupyter Notebook and Embedded.
 */
class Context {
    /**
     * Determines if the given cell contains a notebook tool widget
     *
     * @param cell
     * @returns boolean
     */
    is_tool_cell(cell) {
        const dom_node = cell.node || cell.element;
        if (!!dom_node)
            return !!dom_node.querySelector('.nbtools');
        else
            return false;
    }
}
/**
 * Context handler for JupyterLab
 */
class LabContext extends Context {
    /**
     * Toggle hiding or or showing the code for the specified cell
     *
     * @param {HTMLElement} element
     * @param {boolean} display
     */
    toggle_code(element, display) {
        setTimeout(() => {
            let input_block = element.closest('.jp-Cell');
            if (input_block)
                input_block = input_block.querySelector('.jp-Cell-inputWrapper');
            // Set display to toggle if not specified
            if (!!display)
                utils_1.show(input_block);
            else if (display === false)
                utils_1.hide(input_block);
            else
                utils_1.toggle(input_block);
        }, 100);
    }
    /**
     * Execute the indicated cell in the notebook
     *
     * @param cell
     */
    run_cell(cell = null) {
        if (!ContextManager.notebook_tracker)
            return; // If no notebook_tracker, do nothing
        // If cell is falsy, default to the active cell
        if (!cell)
            cell = ContextManager.notebook_tracker.activeCell;
        // If this is a code cell
        if (cell instanceof cells_1.CodeCell) {
            const current = ContextManager.notebook_tracker.currentWidget;
            if (current) {
                cells_1.CodeCell.execute(cell, current.context.sessionContext);
            }
        }
        // If this is a markdown cell
        // if (ContextManager.notebook_tracker.activeCell instanceof MarkdownCell)
        //     MarkdownCell.renderInput(ContextManager.notebook_tracker.activeCell)
    }
    /**
     * Execute all cells with nbtools widgets in the current notebook
     */
    run_tool_cells() {
        if (!ContextManager.notebook_tracker)
            return; // If no notebook_tracker, do nothing
        if (!ContextManager.notebook_tracker.currentWidget)
            return; // If no notebook selected, do nothing
        const cell_widgets = ContextManager.notebook_tracker.currentWidget.content.widgets;
        cell_widgets.forEach((widget) => {
            if (this.is_tool_cell(widget))
                this.run_cell(widget);
        });
    }
}
/**
 * Context handler for the classic Jupyter Notebook
 */
class NotebookContext extends Context {
    /**
     * Toggle hiding or or showing the code for the specified cell
     *
     * @param {HTMLElement} element
     * @param {boolean} display
     */
    toggle_code(element, display) {
        const code = element.closest(".cell").querySelector(".input");
        // Set display to toggle if not specified
        if (!!display)
            utils_1.show(code);
        else if (display === false)
            utils_1.hide(code);
        else
            utils_1.toggle(code);
    }
    /**
     * Execute the indicated cell in the notebook
     *
     * @param cell
     */
    run_cell(cell) {
        window.Jupyter.notebook.execute(cell);
    }
    /**
     * Execute all cells with nbtools widgets in the current notebook
     */
    run_tool_cells() {
        window.Jupyter.notebook.get_cells().forEach((cell) => {
            if (this.is_tool_cell(cell))
                this.run_cell(cell);
        });
    }
}
/**
 * Context handler for embedded widgets
 */
class EmbedContext extends Context {
    /**
     * No cells in this context, so do nothing
     *
     * @param {HTMLElement} element
     * @param {boolean} display
     */
    toggle_code(element, display) { }
    /**
     * No cells in this context, so do nothing
     * @param cell
     */
    run_cell(cell) { }
    /**
     * No cells in this context, so do nothing
     */
    run_tool_cells() { }
}
//# sourceMappingURL=context.js.map